#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode, TagHttpParams

link_vector = ['https://www.djangoproject.com/']
support_matrix = {
    'django': {
        '>=3.6': ['3.2'],
        # ">=3.8": ["4.0a1"]  # expected Dec 2021
    }
}
note = """"""


def install():
    from django.core.handlers.base import BaseHandler
    from django.core.handlers import exception

    _get_response = BaseHandler.get_response
    _handle_uncaught_exception = exception.handle_uncaught_exception

    def _sw_get_response(this, request):
        if request is None:
            resp = _get_response(this, request)
            return resp

        carrier = Carrier()
        method = request.method

        for item in carrier:
            # Any HTTP headers in the request are converted to META keys by converting all characters to uppercase,
            # replacing any hyphens with underscores and adding an HTTP_ prefix to the name.
            # https://docs.djangoproject.com/en/3.0/ref/request-response/#django.http.HttpRequest.META
            sw_http_header_key = f"HTTP_{item.key.upper().replace('-', '_')}"
            if sw_http_header_key in request.META:
                item.val = request.META[sw_http_header_key]

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=request.path, carrier=carrier)

        with span:
            span.layer = Layer.Http
            span.component = Component.Django
            span.peer = f"{request.META.get('REMOTE_ADDR')}:{request.META.get('REMOTE_PORT') or '80'}"

            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(request.build_absolute_uri().split('?')[0]))

            # you can get request parameters by `request.GET` even though client are using POST or other methods
            if config.django_collect_http_params and request.GET:
                span.tag(TagHttpParams(params_tostring(request.GET)[0:config.http_params_length_threshold]))

            resp = _get_response(this, request)
            span.tag(TagHttpStatusCode(resp.status_code))
            if resp.status_code >= 400:
                span.error_occurred = True
            return resp

    def _sw_handle_uncaught_exception(request, resolver, exc_info):
        if exc_info is not None:
            entry_span = get_context().active_span()
            if entry_span is not None:
                entry_span.raised()

        return _handle_uncaught_exception(request, resolver, exc_info)

    BaseHandler.get_response = _sw_get_response
    exception.handle_uncaught_exception = _sw_handle_uncaught_exception


def params_tostring(params):
    return '\n'.join([f"{k}=[{','.join(params.getlist(k))}]" for k, _ in params.items()])



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbInstance, TagDbStatement, TagDbSqlParameters

link_vector = ['https://www.psycopg.org/']
support_matrix = {
    'psycopg[binary]': {
        '>=3.6': ['3.0']  # psycopg is psycopg3
    }
}
note = """"""


def install_sync():
    import wrapt  # psycopg is read-only C extension objects so they need to be proxied
    import psycopg

    # synchronous

    class ProxyCursor(wrapt.ObjectProxy):
        def __init__(self, cur):
            wrapt.ObjectProxy.__init__(self, cur)

            self._self_cur = cur

        def __enter__(self):
            return ProxyCursor(wrapt.ObjectProxy.__enter__(self))

        def execute(self, query, vars=None, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/execute', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and vars is not None:
                    text = ','.join(str(v) for v in vars)

                    if len(text) > config.sql_parameters_length:
                        text = f'{text[:config.sql_parameters_length]}...'

                    span.tag(TagDbSqlParameters(f'[{text}]'))

                return self._self_cur.execute(query, vars, *args, **kwargs)

        def executemany(self, query, vars_list, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/executemany', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length:
                    max_len = config.sql_parameters_length
                    total_len = 0
                    text_list = []

                    for vars in vars_list:
                        text = f"[{','.join(str(v) for v in vars)}]"
                        total_len += len(text)

                        if total_len > max_len:
                            text_list.append(f'{text[:max_len - total_len]}...')

                            break

                        text_list.append(text)

                    span.tag(TagDbSqlParameters(f"[{','.join(text_list)}]"))

                return self._self_cur.executemany(query, vars_list, *args, **kwargs)

        def stream(self, query, vars=None, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/stream', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and vars is not None:
                    text = ','.join(str(v) for v in vars)

                    if len(text) > config.sql_parameters_length:
                        text = f'{text[:config.sql_parameters_length]}...'

                    span.tag(TagDbSqlParameters(f'[{text}]'))

                yield from self._self_cur.stream(query, vars, *args, **kwargs)

    class ProxyConnection(wrapt.ObjectProxy):
        def __init__(self, conn):
            wrapt.ObjectProxy.__init__(self, conn)

            self._self_conn = conn

        def __enter__(self):
            return ProxyConnection(wrapt.ObjectProxy.__enter__(self))

        def cursor(self, *args, **kwargs):
            return ProxyCursor(self._self_conn.cursor(*args, **kwargs))

    def connect(*args, **kwargs):
        return ProxyConnection(_connect(*args, **kwargs))

    _connect = psycopg.Connection.connect
    psycopg.Connection.connect = connect
    psycopg.connect = connect


def install_async():
    import wrapt  # psycopg is read-only C extension objects so they need to be proxied
    import psycopg

    class ProxyAsyncCursor(wrapt.ObjectProxy):
        def __init__(self, cur):
            wrapt.ObjectProxy.__init__(self, cur)

            self._self_cur = cur

        async def __aenter__(self):
            return self  # return ProxyAsyncCursor(await self._self_cur.__aenter__())  # may be necessary in future version of psycopg if __aenter__() does not return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            return await self._self_cur.__aexit__(exc_type, exc_val, exc_tb)

        async def execute(self, query, vars=None, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/execute', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and vars is not None:
                    text = ','.join(str(v) for v in vars)

                    if len(text) > config.sql_parameters_length:
                        text = f'{text[:config.sql_parameters_length]}...'

                    span.tag(TagDbSqlParameters(f'[{text}]'))

                return await self._self_cur.execute(query, vars, *args, **kwargs)

        async def executemany(self, query, vars_list, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/executemany', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length:
                    max_len = config.sql_parameters_length
                    total_len = 0
                    text_list = []

                    for vars in vars_list:
                        text = f"[{','.join(str(v) for v in vars)}]"
                        total_len += len(text)

                        if total_len > max_len:
                            text_list.append(f'{text[:max_len - total_len]}...')

                            break

                        text_list.append(text)

                    span.tag(TagDbSqlParameters(f"[{','.join(text_list)}]"))

                return await self._self_cur.executemany(query, vars_list, *args, **kwargs)

        async def stream(self, query, vars=None, *args, **kwargs):
            dsn = self.connection.info.get_parameters()
            port = self.connection.info.port
            peer = f"{dsn['host']}:{str(port)}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/stream', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and vars is not None:
                    text = ','.join(str(v) for v in vars)

                    if len(text) > config.sql_parameters_length:
                        text = f'{text[:config.sql_parameters_length]}...'

                    span.tag(TagDbSqlParameters(f'[{text}]'))

                async for r in self._self_cur.stream(query, vars, *args, **kwargs):
                    yield r

    class ProxyAsyncConnection(wrapt.ObjectProxy):
        def __init__(self, conn):
            wrapt.ObjectProxy.__init__(self, conn)

            self._self_conn = conn

        async def __aenter__(self):
            return self  # return ProxyAsyncConnection(await self._self_conn.__aenter__())  # may be necessary in future version of psycopg if __aenter__() does not return self

        async def __aexit__(self, exc_type, exc_val, exc_tb):
            return await self._self_conn.__aexit__(exc_type, exc_val, exc_tb)

        def cursor(self, *args, **kwargs):
            return ProxyAsyncCursor(self._self_conn.cursor(*args, **kwargs))

    async def aconnect(*args, **kwargs):
        return ProxyAsyncConnection(await _aconnect(*args, **kwargs))

    _aconnect = psycopg.AsyncConnection.connect
    psycopg.AsyncConnection.connect = aconnect


def install():
    install_sync()
    install_async()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from urllib.request import Request

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

link_vector = ['https://docs.python.org/3/library/urllib.request.html']
support_matrix = {
    'urllib_request': {
        '>=3.6': ['*']
    }
}
note = """"""


def install():
    import socket
    from urllib.request import OpenerDirector
    from urllib.error import HTTPError

    _open = OpenerDirector.open

    def _sw_open(this: OpenerDirector, fullurl, data=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        if isinstance(fullurl, str):
            fullurl = Request(fullurl, data)

        url = fullurl.selector.split('?')[0] if fullurl.selector else '/'
        method = getattr(fullurl, 'method', None) or ('GET' if data is None else 'POST')

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_exit_span(op=url, peer=fullurl.host, component=Component.General)

        with span:
            carrier = span.inject()
            span.layer = Layer.Http
            code = None

            for item in carrier:
                fullurl.add_header(item.key, item.val)

            try:
                res = _open(this, fullurl, data, timeout)
                code = res.code
            except HTTPError as e:
                code = e.code
                raise
            finally:  # we do this here because it may change in _open()
                span.tag(TagHttpMethod(method))
                span.tag(TagHttpURL(fullurl.full_url))

                if code is not None:
                    span.tag(TagHttpStatusCode(code))

                    if code >= 400:
                        span.error_occurred = True

            return res

    OpenerDirector.open = _sw_open



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking.command.executors.command_executor import CommandExecutor
from skywalking.command.profile_task_command import ProfileTaskCommand
from skywalking import profile
from skywalking.profile.profile_task import ProfileTask


class ProfileTaskCommandExecutor(CommandExecutor):

    def execute(self, command: ProfileTaskCommand):
        profile_task = ProfileTask(task_id=command.task_id,
                                   first_span_op_name=command.endpoint_name,
                                   duration=command.duration,
                                   min_duration_threshold=command.min_duration_threshold,
                                   thread_dump_period=command.dump_period,
                                   max_sampling_count=command.max_sampling_count,
                                   start_time=command.start_time,
                                   create_time=command.create_time)

        profile.profile_task_execution_service.add_profile_task(profile_task)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

link_vector = ['https://trypyramid.com']
support_matrix = {
    'pyramid': {
        '>=3.6': ['1.10', '2.0']
    }
}
note = """"""


def install():
    from pyramid.router import Router

    def _sw_invoke_request(self, request, *args, **kwargs):
        carrier = Carrier()
        method = request.method

        for item in carrier:
            val = request.headers.get(item.key)

            if val is not None:
                item.val = val

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=request.path, carrier=carrier)

        with span:
            span.layer = Layer.Http
            span.component = Component.Pyramid
            span.peer = request.remote_host or request.remote_addr

            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(str(request.url)))

            resp = _invoke_request(self, request, *args, **kwargs)

            span.tag(TagHttpStatusCode(resp.status_code))

            if resp.status_code >= 400:
                span.error_occurred = True

        return resp

    _invoke_request = Router.invoke_request
    Router.invoke_request = _sw_invoke_request



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading


class AtomicRef:
    def __init__(self, var):
        self._lock = threading.Lock()
        self._var = var

    def get(self):
        with self._lock:
            return self._var

    def set(self, new_var):
        with self._lock:
            self._var = new_var

    def compare_and_set(self, expect, update) -> bool:
        """
        Atomically sets the value to the given updated value if the current value == the expected value

        :return: return True if success, False if the actual value was not equal to the expected value.
        """
        with self._lock:
            if self._var == expect:
                self._var = update
                return True
            return False



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" This version of sitecustomize will
1. initializes the SkyWalking Python Agent.
2. invoke an existing sitecustomize.py.
Not compatible with Python <= 3.3

When executing commands with `sw-python run command`
This particular sitecustomize module will be picked up by any valid replacement
process(command) invoked by os.execvp in runner.py, thus will be seen by potentially
incompatible Python versions and system interpreters, which certainly will cause problems.
Therefore, an attempt to use command leading to a different Python interpreter should be stopped.
"""
import importlib
import logging
import os
import platform
import sys


def _get_sw_loader_logger():
    """ Setup a new logger with passed skywalking CLI env vars,
    don't import from skywalking, it may not be on sys.path
    if user misuses the CLI to run programs out of scope
    """
    from logging import getLogger
    logger = getLogger('skywalking-loader')
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s [%(threadName)s] [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.propagate = False
    if os.environ.get('SW_PYTHON_CLI_DEBUG_ENABLED') == 'True':  # set from the original CLI runner
        logger.setLevel(level=logging.DEBUG)
    return logger


_sw_loader_logger = _get_sw_loader_logger()
_sw_loader_logger_debug_enabled = _sw_loader_logger.isEnabledFor(logging.DEBUG)
# DEBUG messages in case execution goes wrong
if _sw_loader_logger_debug_enabled:
    _sw_loader_logger.debug('---------------sitecustomize.py---------------')
    _sw_loader_logger.debug(f'Successfully imported sitecustomize.py from `{__file__}`')
    _sw_loader_logger.debug(f'You are inside working dir - {os.getcwd()}')
    _sw_loader_logger.debug(f'Using Python version - {sys.version} ')
    _sw_loader_logger.debug(f'Using executable at - {sys.executable}')
    _sw_loader_logger.debug(f'System Base Python executable location {sys.base_prefix}')

if sys.prefix != sys.base_prefix:
    _sw_loader_logger.debug('[The SkyWalking agent bootstrapper is running inside a virtual environment]')

# It is possible that someone else also has a sitecustomize.py
# in sys.path either set by .pth files or manually, we need to run them as well
# The path to user sitecustomize.py is after ours in sys.path, needs to be imported here.
# just to be safe, remove loader thoroughly from sys.path and also sys.modules

cleared_path = [p for p in sys.path if p != os.path.dirname(__file__)]
sys.path = cleared_path  # remove this version from path
loaded = sys.modules.pop('sitecustomize', None)  # pop sitecustomize from loaded

# now try to find the original sitecustomize provided in user env
try:
    loaded = importlib.import_module('sitecustomie')
    _sw_loader_logger.debug(f'Found user sitecustomize file {loaded}, imported')
except ImportError:  # ModuleNotFoundError
    _sw_loader_logger.debug('Original sitecustomize module not found, skipping.')
finally:  # surprise the import error by adding loaded back
    sys.modules['sitecustomize'] = loaded

# This sitecustomize by default doesn't remove the loader dir from PYTHONPATH,
# Thus, subprocesses and multiprocessing also inherits this sitecustomize.py
# This behavior can be turned off using a user provided env below
# os.environ['SW_PYTHON_BOOTSTRAP_PROPAGATE']

if os.environ.get('SW_PYTHON_BOOTSTRAP_PROPAGATE') == 'False':
    if os.environ.get('PYTHONPATH'):
        partitioned = os.environ['PYTHONPATH'].split(os.path.pathsep)
        loader_path = os.path.dirname(__file__)
        if loader_path in partitioned:  # check if we are already removed by a third-party
            partitioned.remove(loader_path)
            os.environ['PYTHONPATH'] = os.path.pathsep.join(partitioned)
            _sw_loader_logger.debug('Removed loader from PYTHONPATH, spawned process will not have agent enabled')

# Note that users could be misusing the CLI to call a Python program that
# their Python env doesn't have SkyWalking installed. Or even call another
# env that has another version of SkyWalking installed, which is leads to problems.
# Even if `import skywalking` was successful, doesn't mean its the same one where sw-python CLI was run.
# Needs further checking for Python versions and prefix that passed down from CLI in os.environ.

cli_python_version = os.environ.get('SW_PYTHON_VERSION')
cli_python_prefix = os.environ.get('SW_PYTHON_PREFIX')

version_match = cli_python_version == platform.python_version()

# windows can sometimes make capitalized path, lower them can help
prefix_match = cli_python_prefix.lower() == os.path.realpath(os.path.normpath(sys.prefix)).lower()

if not (version_match and prefix_match):

    _sw_loader_logger.error(
        f'\nPython used by sw-python CLI - v{cli_python_version} at {cli_python_prefix}\n'
        f'Python used by your actual program - v{platform.python_version()} '
        f'at {os.path.realpath(os.path.normpath(sys.prefix))}'
    )
    _sw_loader_logger.error('The sw-python CLI was instructed to run a program '
                            'using an different Python installation '
                            'this is not safe and loader will not proceed. '
                            'Please make sure that sw-python cli, skywalking agent and your '
                            'application are using the same Python installation, '
                            'Rerun with debug flag, `sw-python -d run yourapp` for some troubleshooting information.'
                            'use `which sw-python` to find out the invoked CLI location')
    os._exit(1)  # do not go further

else:
    from skywalking import agent, config

    # also override debug for skywalking agent itself
    if os.environ.get('SW_PYTHON_CLI_DEBUG_ENABLED') == 'True':  # set from the original CLI runner
        config.logging_level = 'DEBUG'

    # Currently supports configs read from os.environ

    # This is for python 3.6 - 3.7(maybe) argv is not set for embedded interpreter thus will cause failure in
    # those libs that imports argv from sys, we need to set it manually if it's not there already
    # otherwise the plugin install will fail and things won't work properly, example - Sanic
    if not hasattr(sys, 'argv'):
        sys.argv = ['']

    # noinspection PyBroadException
    try:
        _sw_loader_logger.debug('SkyWalking Python Agent starting, loader finished.')
        agent.start()
    except Exception:
        _sw_loader_logger.exception('SkyWalking Python Agent failed to start, please inspect your package installation')



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from queue import Queue, Empty
from time import time

from skywalking import config
from skywalking.agent import Protocol
from skywalking.client.http import HttpServiceManagementClient, HttpTraceSegmentReportService, HttpLogDataReportService
from skywalking.loggings import logger, logger_debug_enabled
from skywalking.protocol.logging.Logging_pb2 import LogData
from skywalking.trace.segment import Segment


class HttpProtocol(Protocol):
    def __init__(self):
        self.properties_sent = False
        self.service_management = HttpServiceManagementClient()
        self.traces_reporter = HttpTraceSegmentReportService()
        self.log_reporter = HttpLogDataReportService()

    def fork_after_in_child(self):
        self.service_management.fork_after_in_child()
        self.traces_reporter.fork_after_in_child()

    def heartbeat(self):
        if not self.properties_sent:
            self.service_management.send_instance_props()
            self.properties_sent = True
        self.service_management.send_heart_beat()

    def report(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    segment = queue.get(block=block, timeout=timeout)  # type: Segment
                except Empty:
                    return

                queue.task_done()

                if logger_debug_enabled:
                    logger.debug('reporting segment %s', segment)

                yield segment

        try:
            self.traces_reporter.report(generator=generator())
        except Exception:
            pass

    def report_log(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    log_data = queue.get(block=block, timeout=timeout)  # type: LogData
                except Empty:
                    return
                queue.task_done()
                if logger_debug_enabled:
                    logger.debug('Reporting Log')

                yield log_data

        try:
            self.log_reporter.report(generator=generator())
        except Exception:
            pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/Meter.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1alanguage-agent/Meter.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"$\n\x05Label\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"M\n\x10MeterBucketValue\x12\x0e\n\x06\x62ucket\x18\x01 \x01(\x01\x12\r\n\x05\x63ount\x18\x02 \x01(\x03\x12\x1a\n\x12isNegativeInfinity\x18\x03 \x01(\x08\"U\n\x10MeterSingleValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x06labels\x18\x02 \x03(\x0b\x32\x14.skywalking.v3.Label\x12\r\n\x05value\x18\x03 \x01(\x01\"u\n\x0eMeterHistogram\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\x06labels\x18\x02 \x03(\x0b\x32\x14.skywalking.v3.Label\x12/\n\x06values\x18\x03 \x03(\x0b\x32\x1f.skywalking.v3.MeterBucketValue\"\xbe\x01\n\tMeterData\x12\x36\n\x0bsingleValue\x18\x01 \x01(\x0b\x32\x1f.skywalking.v3.MeterSingleValueH\x00\x12\x32\n\thistogram\x18\x02 \x01(\x0b\x32\x1d.skywalking.v3.MeterHistogramH\x00\x12\x0f\n\x07service\x18\x03 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x04 \x01(\t\x12\x11\n\ttimestamp\x18\x05 \x01(\x03\x42\x08\n\x06metric\"B\n\x13MeterDataCollection\x12+\n\tmeterData\x18\x01 \x03(\x0b\x32\x18.skywalking.v3.MeterData2\xa7\x01\n\x12MeterReportService\x12@\n\x07\x63ollect\x12\x18.skywalking.v3.MeterData\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x12O\n\x0c\x63ollectBatch\x12\".skywalking.v3.MeterDataCollection\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x42s\n3org.apache.skywalking.apm.network.language.agent.v3P\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3b\x06proto3')



_LABEL = DESCRIPTOR.message_types_by_name['Label']
_METERBUCKETVALUE = DESCRIPTOR.message_types_by_name['MeterBucketValue']
_METERSINGLEVALUE = DESCRIPTOR.message_types_by_name['MeterSingleValue']
_METERHISTOGRAM = DESCRIPTOR.message_types_by_name['MeterHistogram']
_METERDATA = DESCRIPTOR.message_types_by_name['MeterData']
_METERDATACOLLECTION = DESCRIPTOR.message_types_by_name['MeterDataCollection']
Label = _reflection.GeneratedProtocolMessageType('Label', (_message.Message,), {
  'DESCRIPTOR' : _LABEL,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Label)
  })
_sym_db.RegisterMessage(Label)

MeterBucketValue = _reflection.GeneratedProtocolMessageType('MeterBucketValue', (_message.Message,), {
  'DESCRIPTOR' : _METERBUCKETVALUE,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeterBucketValue)
  })
_sym_db.RegisterMessage(MeterBucketValue)

MeterSingleValue = _reflection.GeneratedProtocolMessageType('MeterSingleValue', (_message.Message,), {
  'DESCRIPTOR' : _METERSINGLEVALUE,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeterSingleValue)
  })
_sym_db.RegisterMessage(MeterSingleValue)

MeterHistogram = _reflection.GeneratedProtocolMessageType('MeterHistogram', (_message.Message,), {
  'DESCRIPTOR' : _METERHISTOGRAM,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeterHistogram)
  })
_sym_db.RegisterMessage(MeterHistogram)

MeterData = _reflection.GeneratedProtocolMessageType('MeterData', (_message.Message,), {
  'DESCRIPTOR' : _METERDATA,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeterData)
  })
_sym_db.RegisterMessage(MeterData)

MeterDataCollection = _reflection.GeneratedProtocolMessageType('MeterDataCollection', (_message.Message,), {
  'DESCRIPTOR' : _METERDATACOLLECTION,
  '__module__' : 'language_agent.Meter_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.MeterDataCollection)
  })
_sym_db.RegisterMessage(MeterDataCollection)

_METERREPORTSERVICE = DESCRIPTOR.services_by_name['MeterReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n3org.apache.skywalking.apm.network.language.agent.v3P\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3'
  _LABEL._serialized_start=66
  _LABEL._serialized_end=102
  _METERBUCKETVALUE._serialized_start=104
  _METERBUCKETVALUE._serialized_end=181
  _METERSINGLEVALUE._serialized_start=183
  _METERSINGLEVALUE._serialized_end=268
  _METERHISTOGRAM._serialized_start=270
  _METERHISTOGRAM._serialized_end=387
  _METERDATA._serialized_start=390
  _METERDATA._serialized_end=580
  _METERDATACOLLECTION._serialized_start=582
  _METERDATACOLLECTION._serialized_end=648
  _METERREPORTSERVICE._serialized_start=651
  _METERREPORTSERVICE._serialized_end=818
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" User application command runner """
import logging
import os
import platform
import sys
from typing import List

from skywalking.bootstrap import cli_logger
from skywalking.bootstrap.cli import SWRunnerFailure


def execute(command: List[str]) -> None:
    """ Set up environ and invokes the given command to replace current process """

    cli_logger.debug(f'SkyWalking Python agent `runner` received command {command}')

    cli_logger.debug('Adding sitecustomize.py to PYTHONPATH')

    from skywalking.bootstrap.loader import __file__ as loader_dir

    loader_path = os.path.dirname(loader_dir)
    new_path = ''

    python_path = os.environ.get('PYTHONPATH')
    if python_path:  # If there is already a different PYTHONPATH, PREPEND to it as we must get loaded first.
        partitioned = python_path.split(os.path.pathsep)
        if loader_path not in partitioned:  # check if we are already there
            new_path = os.path.pathsep.join([loader_path, python_path])  # type: str

    # When constructing sys.path PYTHONPATH is always
    # before other paths and after interpreter invoker path, which is here or none
    os.environ['PYTHONPATH'] = new_path if new_path else loader_path
    cli_logger.debug(f"Updated PYTHONPATH - {os.environ['PYTHONPATH']}")

    # Used in sitecustomize to compare command's Python installation with CLI
    # If not match, need to stop agent from loading, and kill the process
    os.environ['SW_PYTHON_PREFIX'] = os.path.realpath(os.path.normpath(sys.prefix))
    os.environ['SW_PYTHON_VERSION'] = platform.python_version()

    # Pass down the logger debug setting to the replaced process, need a new logger there
    os.environ['SW_PYTHON_CLI_DEBUG_ENABLED'] = 'True' if cli_logger.level == logging.DEBUG else 'False'

    try:
        cli_logger.debug(f'New process starting with file - `{command[0]}` args - `{command}`')
        os.execvp(command[0], command)
    except OSError:
        raise SWRunnerFailure



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from skywalking.protocol.common.Common_pb2 import KeyStringValuePair
from skywalking.protocol.logging.Logging_pb2 import LogData, LogDataBody, TraceContext, LogTags, TextLog

from skywalking import config, agent
from skywalking.trace.context import get_context
from skywalking.utils.filter import sw_traceback, sw_filter


def install():
    from logging import Logger

    layout = config.log_reporter_layout  # type: str
    if layout:
        from skywalking.log.formatter import SWFormatter
        formatter = SWFormatter(fmt=layout, tb_limit=config.cause_exception_depth)

    _handle = Logger.handle
    log_reporter_level = logging.getLevelName(config.log_reporter_level)  # type: int

    def _sw_handle(self, record):
        if record.name in ['skywalking', 'skywalking-cli', 'skywalking-loader']:  # Ignore SkyWalking internal loggers
            return _handle(self, record)

        if record.levelno < log_reporter_level:
            return _handle(self, record)

        if not config.log_reporter_ignore_filter and not self.filter(record):  # ignore filtered logs
            return _handle(self, record)  # return handle to original if record is vetoed, just to be safe

        def build_log_tags() -> LogTags:
            core_tags = [
                KeyStringValuePair(key='level', value=str(record.levelname)),
                KeyStringValuePair(key='logger', value=str(record.name)),
                KeyStringValuePair(key='thread', value=str(record.threadName))
            ]
            l_tags = LogTags()
            l_tags.data.extend(core_tags)

            if config.log_reporter_formatted:
                return l_tags

            for i, arg in enumerate(record.args):
                l_tags.data.append(KeyStringValuePair(key=f'argument.{str(i)}', value=str(arg)))

            if record.exc_info:
                l_tags.data.append(KeyStringValuePair(key='exception',
                                                      value=sw_traceback()
                                                      ))  # \n doesn't work in tags for UI
            return l_tags

        context = get_context()

        log_data = LogData(
            timestamp=round(record.created * 1000),
            service=config.service_name,
            serviceInstance=config.service_instance,
            body=LogDataBody(
                type='text',
                text=TextLog(
                    text=sw_filter(transform(record))
                )
            ),
            traceContext=TraceContext(
                traceId=str(context.segment.related_traces[0]),
                traceSegmentId=str(context.segment.segment_id),
                spanId=context.active_span().sid if context.active_span() else -1
            ),
            tags=build_log_tags(),
        )
        _handle(self=self, record=record)

        agent.archive_log(log_data)

    Logger.handle = _sw_handle

    def transform(record) -> str:
        if config.log_reporter_formatted:
            if layout:
                return formatter.format(record=record)
            newline = '\n'
            return f"{record.getMessage()}{f'{newline}{sw_traceback()}' if record.exc_info else ''}"
        return str(record.msg)  # convert possible exception to str



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from skywalking.trace.context import SpanContext

from skywalking.trace import ID


class Snapshot:
    def __init__(
            self,
            segment_id: str = None,
            span_id: int = None,
            trace_id: ID = None,
            endpoint: str = None,
            correlation: dict = None
    ):
        self.trace_id = trace_id  # type: ID
        self.segment_id = segment_id  # type: str
        self.span_id = span_id  # type: int
        self.endpoint = endpoint  # type: str
        self.correlation = correlation.copy()  # type: dict

    def is_from_current(self, context: 'SpanContext'):
        return self.segment_id is not None and self.segment_id == context.capture().segment_id

    def is_valid(self):
        return self.segment_id is not None and self.span_id > -1 and self.trace_id is not None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/Tracing.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1clanguage-agent/Tracing.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\xa3\x01\n\rSegmentObject\x12\x0f\n\x07traceId\x18\x01 \x01(\t\x12\x16\n\x0etraceSegmentId\x18\x02 \x01(\t\x12(\n\x05spans\x18\x03 \x03(\x0b\x32\x19.skywalking.v3.SpanObject\x12\x0f\n\x07service\x18\x04 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x05 \x01(\t\x12\x15\n\risSizeLimited\x18\x06 \x01(\x08\"\xf0\x01\n\x10SegmentReference\x12\'\n\x07refType\x18\x01 \x01(\x0e\x32\x16.skywalking.v3.RefType\x12\x0f\n\x07traceId\x18\x02 \x01(\t\x12\x1c\n\x14parentTraceSegmentId\x18\x03 \x01(\t\x12\x14\n\x0cparentSpanId\x18\x04 \x01(\x05\x12\x15\n\rparentService\x18\x05 \x01(\t\x12\x1d\n\x15parentServiceInstance\x18\x06 \x01(\t\x12\x16\n\x0eparentEndpoint\x18\x07 \x01(\t\x12 \n\x18networkAddressUsedAtPeer\x18\x08 \x01(\t\"\x91\x03\n\nSpanObject\x12\x0e\n\x06spanId\x18\x01 \x01(\x05\x12\x14\n\x0cparentSpanId\x18\x02 \x01(\x05\x12\x11\n\tstartTime\x18\x03 \x01(\x03\x12\x0f\n\x07\x65ndTime\x18\x04 \x01(\x03\x12-\n\x04refs\x18\x05 \x03(\x0b\x32\x1f.skywalking.v3.SegmentReference\x12\x15\n\roperationName\x18\x06 \x01(\t\x12\x0c\n\x04peer\x18\x07 \x01(\t\x12)\n\x08spanType\x18\x08 \x01(\x0e\x32\x17.skywalking.v3.SpanType\x12+\n\tspanLayer\x18\t \x01(\x0e\x32\x18.skywalking.v3.SpanLayer\x12\x13\n\x0b\x63omponentId\x18\n \x01(\x05\x12\x0f\n\x07isError\x18\x0b \x01(\x08\x12/\n\x04tags\x18\x0c \x03(\x0b\x32!.skywalking.v3.KeyStringValuePair\x12 \n\x04logs\x18\r \x03(\x0b\x32\x12.skywalking.v3.Log\x12\x14\n\x0cskipAnalysis\x18\x0e \x01(\x08\"D\n\x03Log\x12\x0c\n\x04time\x18\x01 \x01(\x03\x12/\n\x04\x64\x61ta\x18\x02 \x03(\x0b\x32!.skywalking.v3.KeyStringValuePair\"\x10\n\x02ID\x12\n\n\x02id\x18\x01 \x03(\t\"C\n\x11SegmentCollection\x12.\n\x08segments\x18\x01 \x03(\x0b\x32\x1c.skywalking.v3.SegmentObject**\n\x08SpanType\x12\t\n\x05\x45ntry\x10\x00\x12\x08\n\x04\x45xit\x10\x01\x12\t\n\x05Local\x10\x02*,\n\x07RefType\x12\x10\n\x0c\x43rossProcess\x10\x00\x12\x0f\n\x0b\x43rossThread\x10\x01*_\n\tSpanLayer\x12\x0b\n\x07Unknown\x10\x00\x12\x0c\n\x08\x44\x61tabase\x10\x01\x12\x10\n\x0cRPCFramework\x10\x02\x12\x08\n\x04Http\x10\x03\x12\x06\n\x02MQ\x10\x04\x12\t\n\x05\x43\x61\x63he\x10\x05\x12\x08\n\x04\x46\x41\x41S\x10\x06\x32\xaf\x01\n\x19TraceSegmentReportService\x12\x44\n\x07\x63ollect\x12\x1c.skywalking.v3.SegmentObject\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x12L\n\rcollectInSync\x12 .skywalking.v3.SegmentCollection\x1a\x17.skywalking.v3.Commands\"\x00\x42\x93\x01\n3org.apache.skywalking.apm.network.language.agent.v3P\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')

_SPANTYPE = DESCRIPTOR.enum_types_by_name['SpanType']
SpanType = enum_type_wrapper.EnumTypeWrapper(_SPANTYPE)
_REFTYPE = DESCRIPTOR.enum_types_by_name['RefType']
RefType = enum_type_wrapper.EnumTypeWrapper(_REFTYPE)
_SPANLAYER = DESCRIPTOR.enum_types_by_name['SpanLayer']
SpanLayer = enum_type_wrapper.EnumTypeWrapper(_SPANLAYER)
Entry = 0
Exit = 1
Local = 2
CrossProcess = 0
CrossThread = 1
Unknown = 0
Database = 1
RPCFramework = 2
Http = 3
MQ = 4
Cache = 5
FAAS = 6


_SEGMENTOBJECT = DESCRIPTOR.message_types_by_name['SegmentObject']
_SEGMENTREFERENCE = DESCRIPTOR.message_types_by_name['SegmentReference']
_SPANOBJECT = DESCRIPTOR.message_types_by_name['SpanObject']
_LOG = DESCRIPTOR.message_types_by_name['Log']
_ID = DESCRIPTOR.message_types_by_name['ID']
_SEGMENTCOLLECTION = DESCRIPTOR.message_types_by_name['SegmentCollection']
SegmentObject = _reflection.GeneratedProtocolMessageType('SegmentObject', (_message.Message,), {
  'DESCRIPTOR' : _SEGMENTOBJECT,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.SegmentObject)
  })
_sym_db.RegisterMessage(SegmentObject)

SegmentReference = _reflection.GeneratedProtocolMessageType('SegmentReference', (_message.Message,), {
  'DESCRIPTOR' : _SEGMENTREFERENCE,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.SegmentReference)
  })
_sym_db.RegisterMessage(SegmentReference)

SpanObject = _reflection.GeneratedProtocolMessageType('SpanObject', (_message.Message,), {
  'DESCRIPTOR' : _SPANOBJECT,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.SpanObject)
  })
_sym_db.RegisterMessage(SpanObject)

Log = _reflection.GeneratedProtocolMessageType('Log', (_message.Message,), {
  'DESCRIPTOR' : _LOG,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Log)
  })
_sym_db.RegisterMessage(Log)

ID = _reflection.GeneratedProtocolMessageType('ID', (_message.Message,), {
  'DESCRIPTOR' : _ID,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.ID)
  })
_sym_db.RegisterMessage(ID)

SegmentCollection = _reflection.GeneratedProtocolMessageType('SegmentCollection', (_message.Message,), {
  'DESCRIPTOR' : _SEGMENTCOLLECTION,
  '__module__' : 'language_agent.Tracing_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.SegmentCollection)
  })
_sym_db.RegisterMessage(SegmentCollection)

_TRACESEGMENTREPORTSERVICE = DESCRIPTOR.services_by_name['TraceSegmentReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n3org.apache.skywalking.apm.network.language.agent.v3P\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _SPANTYPE._serialized_start=1038
  _SPANTYPE._serialized_end=1080
  _REFTYPE._serialized_start=1082
  _REFTYPE._serialized_end=1126
  _SPANLAYER._serialized_start=1128
  _SPANLAYER._serialized_end=1223
  _SEGMENTOBJECT._serialized_start=69
  _SEGMENTOBJECT._serialized_end=232
  _SEGMENTREFERENCE._serialized_start=235
  _SEGMENTREFERENCE._serialized_end=475
  _SPANOBJECT._serialized_start=478
  _SPANOBJECT._serialized_end=879
  _LOG._serialized_start=881
  _LOG._serialized_end=949
  _ID._serialized_start=951
  _ID._serialized_end=967
  _SEGMENTCOLLECTION._serialized_start=969
  _SEGMENTCOLLECTION._serialized_end=1036
  _TRACESEGMENTREPORTSERVICE._serialized_start=1226
  _TRACESEGMENTREPORTSERVICE._serialized_end=1401
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode, TagHttpParams

link_vector = ['https://flask.palletsprojects.com']
support_matrix = {
    'flask': {
        '>=3.6': ['1.1', '2.0']  # 1.1 to be removed in near future
    }
}
note = """"""


def install():
    from flask import Flask
    _full_dispatch_request = Flask.full_dispatch_request
    _handle_user_exception = Flask.handle_user_exception
    _handle_exception = Flask.handle_exception

    def params_tostring(params):
        return '\n'.join([f"{k}=[{','.join(params.getlist(k))}]" for k, _ in params.items()])

    def _sw_full_dispatch_request(this: Flask):
        import flask

        req = flask.request
        carrier = Carrier()
        method = req.method

        for item in carrier:
            if item.key.capitalize() in req.headers:
                item.val = req.headers[item.key.capitalize()]

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_entry_span(op=req.path, carrier=carrier, inherit=Component.General)

        with span:
            span.layer = Layer.Http
            span.component = Component.Flask
            if all(environ_key in req.environ for environ_key in ('REMOTE_ADDR', 'REMOTE_PORT')):
                span.peer = f"{req.environ['REMOTE_ADDR']}:{req.environ['REMOTE_PORT']}"
            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(req.url.split('?')[0]))
            if config.flask_collect_http_params and req.values:
                span.tag(TagHttpParams(params_tostring(req.values)[0:config.http_params_length_threshold]))
            resp = _full_dispatch_request(this)

            if resp.status_code >= 400:
                span.error_occurred = True

            span.tag(TagHttpStatusCode(resp.status_code))
            return resp

    def _sw_handle_user_exception(this: Flask, e):
        if e is not None:
            entry_span = get_context().active_span()
            if entry_span is not None and type(entry_span) is not NoopSpan:
                entry_span.raised()

        return _handle_user_exception(this, e)

    def _sw_handle_exception(this: Flask, e):
        if e is not None:
            entry_span = get_context().active_span()
            if entry_span is not None and type(entry_span) is not NoopSpan:
                entry_span.raised()

        return _handle_exception(this, e)

    Flask.full_dispatch_request = _sw_full_dispatch_request
    Flask.handle_user_exception = _sw_handle_user_exception
    Flask.handle_exception = _sw_handle_exception



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: language-agent/JVMMetricCompat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2
from ..language_agent import JVMMetric_pb2 as language__agent_dot_JVMMetric__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$language-agent/JVMMetricCompat.proto\x1a\x13\x63ommon/Common.proto\x1a\x1elanguage-agent/JVMMetric.proto2b\n\x16JVMMetricReportService\x12H\n\x07\x63ollect\x12\".skywalking.v3.JVMMetricCollection\x1a\x17.skywalking.v3.Commands\"\x00\x42\xa4\x01\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_JVMMETRICREPORTSERVICE = DESCRIPTOR.services_by_name['JVMMetricReportService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _JVMMETRICREPORTSERVICE._serialized_start=93
  _JVMMETRICREPORTSERVICE._serialized_end=191
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from skywalking import config


def getLogger(name=None):  # noqa
    logger = logging.getLogger(name)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s [%(threadName)s] [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.propagate = False

    return logger


logger = getLogger('skywalking')
logger_debug_enabled = logger.isEnabledFor(logging.DEBUG)


def init():
    logging.addLevelName(logging.CRITICAL + 10, 'OFF')
    logger.setLevel(logging.getLevelName(config.logging_level))



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import inspect
import logging
import pkgutil
import re
import traceback

import pkg_resources
from packaging import version

import skywalking
from skywalking import config
from skywalking.loggings import logger
from skywalking.utils.comparator import operators
from skywalking.utils.exception import VersionRuleException


def install():
    disable_patterns = config.disable_plugins
    if isinstance(disable_patterns, str):
        disable_patterns = [re.compile(p.strip()) for p in disable_patterns.split(',') if p.strip()]
    else:
        disable_patterns = [re.compile(p.strip()) for p in disable_patterns if p.strip()]
    for importer, modname, _ispkg in pkgutil.iter_modules(skywalking.plugins.__path__):
        if any(pattern.match(modname) for pattern in disable_patterns):
            logger.info("plugin %s is disabled and thus won't be installed", modname)
            continue
        logger.debug('installing plugin %s', modname)
        plugin = importer.find_module(modname).load_module(modname)

        # todo: refactor the version checker, currently it doesn't really work as intended
        supported = pkg_version_check(plugin)
        if not supported:
            logger.debug("check version for plugin %s's corresponding package failed, thus "
                         "won't be installed", modname)
            continue

        if not hasattr(plugin, 'install') or inspect.ismethod(plugin.install):
            logger.warning("no `install` method in plugin %s, thus the plugin won't be installed", modname)
            continue

        # noinspection PyBroadException
        try:
            plugin.install()
            logger.debug('Successfully installed plugin %s', modname)
        except Exception:
            logger.warning(
                'plugin %s failed to install, please disregard this warning '
                'if the corresponding package was not used in your project',
                modname
            )
            traceback.print_exc() if logger.isEnabledFor(logging.DEBUG) else None


def pkg_version_check(plugin):
    supported = True

    # no version rules was set, no checks
    if not hasattr(plugin, 'version_rule'):
        return supported

    pkg_name = plugin.version_rule.get('name')
    rules = plugin.version_rule.get('rules')

    try:
        current_pkg_version = pkg_resources.get_distribution(pkg_name).version
    except pkg_resources.DistributionNotFound:
        # when failed to get the version, we consider it as supported.
        return supported

    current_version = version.parse(current_pkg_version)
    # pass one rule in rules (OR)
    for rule in rules:
        if rule.find(' ') == -1:
            if check(rule, current_version):
                return supported
        else:
            # have to pass all rule_uint in this rule (AND)
            rule_units = rule.split(' ')
            results = [check(unit, current_version) for unit in rule_units]
            if False in results:
                # check failed, try to check next rule
                continue
            else:
                return supported

    supported = False
    return supported


def check(rule_unit, current_version):
    idx = 2 if rule_unit[1] == '=' else 1
    symbol = rule_unit[0:idx]
    expect_pkg_version = rule_unit[idx:]

    expect_version = version.parse(expect_pkg_version)
    f = operators.get(symbol) or None
    if not f:
        raise VersionRuleException(f'version rule {rule_unit} error. only allow >,>=,==,<=,<,!= symbols')

    return f(current_version, expect_version)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect
from functools import wraps
from typing import List

from skywalking import Layer, Component
from skywalking.trace.context import get_context
from skywalking.trace.tags import Tag


def trace(
        op: str = None,
        layer: Layer = Layer.Unknown,
        component: Component = Component.Unknown,
        tags: List[Tag] = None,
):
    def decorator(func):
        _op = op or func.__name__

        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                context = get_context()
                span = context.new_local_span(op=_op)
                span.layer = layer
                span.component = component
                if tags:
                    for tag in tags:
                        span.tag(tag)
                with span:
                    return await func(*args, **kwargs)

            return wrapper

        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                context = get_context()
                span = context.new_local_span(op=_op)
                span.layer = layer
                span.component = component
                if tags:
                    for tag in tags:
                        span.tag(tag)
                with span:
                    return func(*args, **kwargs)

            return wrapper

    return decorator


def runnable(
        op: str = None,
        layer: Layer = Layer.Unknown,
        component: Component = Component.Unknown,
        tags: List[Tag] = None,
):
    def decorator(func):
        snapshot = get_context().capture()

        @wraps(func)
        def wrapper(*args, **kwargs):
            _op = op or f'Thread/{func.__name__}'
            context = get_context()
            with context.new_local_span(op=_op) as span:
                context.continued(snapshot)
                span.layer = layer
                span.component = component
                if tags:
                    for tag in tags:
                        span.tag(tag)
                func(*args, **kwargs)

        return wrapper

    return decorator



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from ..common import Common_pb2 as common_dot_Common__pb2
from ..language_agent import Tracing_pb2 as language__agent_dot_Tracing__pb2


class TraceSegmentReportServiceStub(object):
    """Define a trace segment report service.
    All language agents or any trace collecting component, could use this service to send span collection to the SkyWalking OAP backend.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.collect = channel.stream_unary(
                '/TraceSegmentReportService/collect',
                request_serializer=language__agent_dot_Tracing__pb2.SegmentObject.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )
        self.collectInSync = channel.unary_unary(
                '/TraceSegmentReportService/collectInSync',
                request_serializer=language__agent_dot_Tracing__pb2.SegmentCollection.SerializeToString,
                response_deserializer=common_dot_Common__pb2.Commands.FromString,
                )


class TraceSegmentReportServiceServicer(object):
    """Define a trace segment report service.
    All language agents or any trace collecting component, could use this service to send span collection to the SkyWalking OAP backend.
    """

    def collect(self, request_iterator, context):
        """Recommended trace segment report channel.
        gRPC streaming provides better performance.
        All language agents should choose this.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def collectInSync(self, request, context):
        """An alternative for trace report by using gRPC unary
        This is provided for some 3rd-party integration, if and only if they prefer the unary mode somehow.
        The performance of SkyWalking OAP server would be very similar with streaming report,
        the performance of the network and client side are affected
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TraceSegmentReportServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'collect': grpc.stream_unary_rpc_method_handler(
                    servicer.collect,
                    request_deserializer=language__agent_dot_Tracing__pb2.SegmentObject.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
            'collectInSync': grpc.unary_unary_rpc_method_handler(
                    servicer.collectInSync,
                    request_deserializer=language__agent_dot_Tracing__pb2.SegmentCollection.FromString,
                    response_serializer=common_dot_Common__pb2.Commands.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'TraceSegmentReportService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TraceSegmentReportService(object):
    """Define a trace segment report service.
    All language agents or any trace collecting component, could use this service to send span collection to the SkyWalking OAP backend.
    """

    @staticmethod
    def collect(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/TraceSegmentReportService/collect',
            language__agent_dot_Tracing__pb2.SegmentObject.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def collectInSync(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/TraceSegmentReportService/collectInSync',
            language__agent_dot_Tracing__pb2.SegmentCollection.SerializeToString,
            common_dot_Common__pb2.Commands.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re
import uuid
from typing import TYPE_CHECKING
import yaml

if TYPE_CHECKING:
    from typing import List

QUEUE_TIMEOUT = 1  # type: int

RE_IGNORE_PATH = re.compile('^$')  # type: re.Pattern
RE_HTTP_IGNORE_METHOD = RE_IGNORE_PATH  # type: re.Pattern

options = None  # here to include 'options' in globals
options = globals().copy()  # THIS MUST PRECEDE DIRECTLY BEFORE LIST OF CONFIG OPTIONS!

yamlPath = os.path.join("/etc", "skywalking-config.yml")
data = None
service_instance = None
trace_ignore_path = None

if os.path.exists(yamlPath):
    f = open(yamlPath, 'r', encoding='utf-8')
    cfg = f.read()
    data = yaml.load(cfg, Loader=yaml.FullLoader)

service_name = os.getenv('SW_AGENT_NAME') or 'Python Service Name'  # type: str
if data:
    service_instance = data.get('service_instance')
if not service_instance:
    service_instance = os.getenv('SW_AGENT_INSTANCE') or str(uuid.uuid1()).replace('-', '')  # type: str

agent_namespace = os.getenv('SW_AGENT_NAMESPACE')  # type: str
collector_address = os.getenv('SW_AGENT_COLLECTOR_BACKEND_SERVICES') or '127.0.0.1:11800'  # type: str
force_tls = os.getenv('SW_AGENT_FORCE_TLS', '').lower() == 'true'  # type: bool
protocol = (os.getenv('SW_AGENT_PROTOCOL') or 'grpc').lower()  # type: str
authentication = os.getenv('SW_AGENT_AUTHENTICATION')  # type: str
logging_level = os.getenv('SW_AGENT_LOGGING_LEVEL') or 'INFO'  # type: str
disable_plugins = (os.getenv('SW_AGENT_DISABLE_PLUGINS') or '').split(',')  # type: List[str]
max_buffer_size = int(os.getenv('SW_AGENT_MAX_BUFFER_SIZE', '10000'))  # type: int
sql_parameters_length = int(os.getenv('SW_SQL_PARAMETERS_LENGTH') or '0')  # type: int
pymongo_trace_parameters = True if os.getenv('SW_PYMONGO_TRACE_PARAMETERS') and \
                                   os.getenv('SW_PYMONGO_TRACE_PARAMETERS') == 'True' else False  # type: bool
pymongo_parameters_max_length = int(os.getenv('SW_PYMONGO_PARAMETERS_MAX_LENGTH') or '512')  # type: int
ignore_suffix = os.getenv('SW_IGNORE_SUFFIX') or '.jpg,.jpeg,.js,.css,.png,.bmp,.gif,.ico,.mp3,' \
                                                 '.mp4,.html,.svg '  # type: str
flask_collect_http_params = True if os.getenv('SW_FLASK_COLLECT_HTTP_PARAMS') and \
                                    os.getenv('SW_FLASK_COLLECT_HTTP_PARAMS') == 'True' else False  # type: bool
sanic_collect_http_params = True if os.getenv('SW_SANIC_COLLECT_HTTP_PARAMS') and \
                                    os.getenv('SW_SANIC_COLLECT_HTTP_PARAMS') == 'True' else False  # type: bool
http_params_length_threshold = int(os.getenv('SW_HTTP_PARAMS_LENGTH_THRESHOLD') or '1024')  # type: int
http_ignore_method = os.getenv('SW_HTTP_IGNORE_METHOD', '').upper()  # type: str
django_collect_http_params = True if os.getenv('SW_DJANGO_COLLECT_HTTP_PARAMS') and \
                                     os.getenv('SW_DJANGO_COLLECT_HTTP_PARAMS') == 'True' else False  # type: bool
correlation_element_max_number = int(os.getenv('SW_CORRELATION_ELEMENT_MAX_NUMBER') or '3')  # type: int
correlation_value_max_length = int(os.getenv('SW_CORRELATION_VALUE_MAX_LENGTH') or '128')  # type: int

if data:
    trace_ignore_path = data.get('trace_ignore_path')
if not trace_ignore_path:
    trace_ignore_path = os.getenv('SW_TRACE_IGNORE_PATH') or ''  # type: str

elasticsearch_trace_dsl = True if os.getenv('SW_ELASTICSEARCH_TRACE_DSL') and \
                                  os.getenv('SW_ELASTICSEARCH_TRACE_DSL') == 'True' else False  # type: bool
kafka_bootstrap_servers = os.getenv('SW_KAFKA_REPORTER_BOOTSTRAP_SERVERS') or 'localhost:9092'  # type: str
kafka_topic_management = os.getenv('SW_KAFKA_REPORTER_TOPIC_MANAGEMENT') or 'skywalking-managements'  # type: str
kafka_topic_segment = os.getenv('SW_KAFKA_REPORTER_TOPIC_SEGMENT') or 'skywalking-segments'  # type: str
kafka_topic_log = os.getenv('SW_KAFKA_REPORTER_TOPIC_LOG') or 'skywalking-logs'  # type: str
celery_parameters_length = int(os.getenv('SW_CELERY_PARAMETERS_LENGTH') or '512')
fastapi_collect_http_params = True if os.getenv('SW_FASTAPI_COLLECT_HTTP_PARAMS') and \
                                      os.getenv('SW_FASTAPI_COLLECT_HTTP_PARAMS') == 'True' else False  # type: bool

# profile configs
get_profile_task_interval = int(os.getenv('SW_PROFILE_TASK_QUERY_INTERVAL') or '20')  # type: int
profile_active = False if os.getenv('SW_AGENT_PROFILE_ACTIVE') and \
                          os.getenv('SW_AGENT_PROFILE_ACTIVE') == 'False' else True  # type: bool
profile_max_parallel = int(os.getenv('SW_AGENT_PROFILE_MAX_PARALLEL') or '5')  # type: int
profile_duration = int(os.getenv('SW_AGENT_PROFILE_DURATION') or '10')  # type: int
profile_dump_max_stack_depth = int(os.getenv('SW_AGENT_PROFILE_DUMP_MAX_STACK_DEPTH') or '500')  # type: int
profile_snapshot_transport_buffer_size = int(os.getenv('SW_AGENT_PROFILE_SNAPSHOT_TRANSPORT_BUFFER_SIZE') or '50')

log_reporter_active = True if os.getenv('SW_AGENT_LOG_REPORTER_ACTIVE') and \
                              os.getenv('SW_AGENT_LOG_REPORTER_ACTIVE') == 'True' else False  # type: bool
log_reporter_max_buffer_size = int(os.getenv('SW_AGENT_LOG_REPORTER_BUFFER_SIZE') or '10000')  # type: int
log_reporter_level = os.getenv('SW_AGENT_LOG_REPORTER_LEVEL') or 'WARNING'  # type: str
log_reporter_ignore_filter = True if os.getenv('SW_AGENT_LOG_REPORTER_IGNORE_FILTER') and \
                                     os.getenv('SW_AGENT_LOG_REPORTER_IGNORE_FILTER') == 'True' else False  # type: bool
log_reporter_formatted = False if os.getenv('SW_AGENT_LOG_REPORTER_FORMATTED') and \
                                  os.getenv('SW_AGENT_LOG_REPORTER_FORMATTED') == 'False' else True  # type: bool
log_reporter_layout = os.getenv('SW_AGENT_LOG_REPORTER_LAYOUT') or \
                      '%(asctime)s [%(threadName)s] %(levelname)s %(name)s - %(message)s'  # type: str
cause_exception_depth = int(os.getenv('SW_AGENT_CAUSE_EXCEPTION_DEPTH') or '5')  # type: int


options = {key for key in globals() if key not in options}  # THIS MUST FOLLOW DIRECTLY AFTER LIST OF CONFIG OPTIONS!


def init(**kwargs):
    glob = globals()

    for key, val in kwargs.items():
        if key not in options:
            raise KeyError(f'invalid config option {key}')

        glob[key] = val


def finalize():
    reesc = re.compile(r'([.*+?^=!:${}()|\[\]\\])')
    suffix = r'^.+(?:' + '|'.join(reesc.sub(r'\\\1', s.strip()) for s in ignore_suffix.split(',')) + ')$'
    method = r'^' + '|'.join(s.strip() for s in http_ignore_method.split(',')) + '$'
    path = '^(?:' + \
           '|'.join(  # replaces ","
               '(?:(?:[^/]+/)*[^/]+)?'.join(  # replaces "**"
                   '[^/]*'.join(  # replaces "*"
                       '[^/]'.join(  # replaces "?"
                           reesc.sub(r'\\\1', s) for s in p2.split('?')
                       ) for p2 in p1.split('*')
                   ) for p1 in p0.strip().split('**')
               ) for p0 in trace_ignore_path.split(',')
           ) + ')$'

    global RE_IGNORE_PATH, RE_HTTP_IGNORE_METHOD
    RE_IGNORE_PATH = re.compile(f'{suffix}|{path}')
    RE_HTTP_IGNORE_METHOD = re.compile(method, re.IGNORECASE)


def ignore_http_method_check(method: str):
    return RE_HTTP_IGNORE_METHOD.match(method)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import traceback
from queue import Queue, Empty
from time import time

import grpc

from skywalking import config
from skywalking.agent import Protocol
from skywalking.agent.protocol.interceptors import header_adder_interceptor
from skywalking.client.grpc import GrpcServiceManagementClient, GrpcTraceSegmentReportService, \
    GrpcProfileTaskChannelService, GrpcLogDataReportService
from skywalking.loggings import logger, logger_debug_enabled
from skywalking.profile.profile_task import ProfileTask
from skywalking.profile.snapshot import TracingThreadSnapshot
from skywalking.protocol.common.Common_pb2 import KeyStringValuePair
from skywalking.protocol.language_agent.Tracing_pb2 import SegmentObject, SpanObject, Log, SegmentReference
from skywalking.protocol.logging.Logging_pb2 import LogData
from skywalking.protocol.profile.Profile_pb2 import ThreadSnapshot, ThreadStack
from skywalking.trace.segment import Segment


class GrpcProtocol(Protocol):
    def __init__(self):
        self.properties_sent = False
        self.state = None

        if config.force_tls:
            self.channel = grpc.secure_channel(config.collector_address, grpc.ssl_channel_credentials())
        else:
            self.channel = grpc.insecure_channel(config.collector_address)

        if config.authentication:
            self.channel = grpc.intercept_channel(
                self.channel, header_adder_interceptor('authentication', config.authentication)
            )

        self.channel.subscribe(self._cb, try_to_connect=True)
        self.service_management = GrpcServiceManagementClient(self.channel)
        self.traces_reporter = GrpcTraceSegmentReportService(self.channel)
        self.profile_channel = GrpcProfileTaskChannelService(self.channel)
        self.log_reporter = GrpcLogDataReportService(self.channel)

    def _cb(self, state):
        if logger_debug_enabled:
            logger.debug('grpc channel connectivity changed, [%s -> %s]', self.state, state)
        self.state = state

    def query_profile_commands(self):
        if logger_debug_enabled:
            logger.debug('query profile commands')
        self.profile_channel.do_query()

    def notify_profile_task_finish(self, task: ProfileTask):
        self.profile_channel.finish(task)

    def heartbeat(self):
        try:
            if not self.properties_sent:
                self.service_management.send_instance_props()
                self.properties_sent = True

            self.service_management.send_heart_beat()

        except grpc.RpcError:
            self.on_error()
            raise

    def on_error(self):
        traceback.print_exc() if logger.isEnabledFor(logging.DEBUG) else None
        self.channel.unsubscribe(self._cb)
        self.channel.subscribe(self._cb, try_to_connect=True)

    def report(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    segment = queue.get(block=block, timeout=timeout)  # type: Segment
                except Empty:
                    return

                queue.task_done()

                if logger_debug_enabled:
                    logger.debug('reporting segment %s', segment)

                s = SegmentObject(
                    traceId=str(segment.related_traces[0]),
                    traceSegmentId=str(segment.segment_id),
                    service=config.service_name,
                    serviceInstance=config.service_instance,
                    spans=[SpanObject(
                        spanId=span.sid,
                        parentSpanId=span.pid,
                        startTime=span.start_time,
                        endTime=span.end_time,
                        operationName=span.op,
                        peer=span.peer,
                        spanType=span.kind.name,
                        spanLayer=span.layer.name,
                        componentId=span.component.value,
                        isError=span.error_occurred,
                        logs=[Log(
                            time=int(log.timestamp * 1000),
                            data=[KeyStringValuePair(key=item.key, value=item.val) for item in log.items],
                        ) for log in span.logs],
                        tags=[KeyStringValuePair(
                            key=tag.key,
                            value=str(tag.val),
                        ) for tag in span.iter_tags()],
                        refs=[SegmentReference(
                            refType=0 if ref.ref_type == 'CrossProcess' else 1,
                            traceId=ref.trace_id,
                            parentTraceSegmentId=ref.segment_id,
                            parentSpanId=ref.span_id,
                            parentService=ref.service,
                            parentServiceInstance=ref.service_instance,
                            parentEndpoint=ref.endpoint,
                            networkAddressUsedAtPeer=ref.client_address,
                        ) for ref in span.refs if ref.trace_id],
                    ) for span in segment.spans],
                )

                yield s

        try:
            self.traces_reporter.report(generator())
        except grpc.RpcError:
            self.on_error()
            raise  # reraise so that incremental reconnect wait can process

    def report_log(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    log_data = queue.get(block=block, timeout=timeout)  # type: LogData
                except Empty:
                    return

                queue.task_done()

                if logger_debug_enabled:
                    logger.debug('Reporting Log')

                yield log_data

        try:
            self.log_reporter.report(generator())
        except grpc.RpcError:
            self.on_error()
            raise

    def send_snapshot(self, queue: Queue, block: bool = True):
        start = None

        def generator():
            nonlocal start

            while True:
                try:
                    timeout = config.QUEUE_TIMEOUT  # type: int
                    if not start:  # make sure first time through queue is always checked
                        start = time()
                    else:
                        timeout -= int(time() - start)
                        if timeout <= 0:  # this is to make sure we exit eventually instead of being fed continuously
                            return
                    snapshot = queue.get(block=block, timeout=timeout)  # type: TracingThreadSnapshot
                except Empty:
                    return

                queue.task_done()

                transform_snapshot = ThreadSnapshot(
                    taskId=str(snapshot.task_id),
                    traceSegmentId=str(snapshot.trace_segment_id),
                    time=int(snapshot.time),
                    sequence=int(snapshot.sequence),
                    stack=ThreadStack(codeSignatures=snapshot.stack_list)
                )

                yield transform_snapshot

        try:
            self.profile_channel.send(generator())
        except grpc.RpcError:
            self.on_error()
            raise



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbInstance, TagDbStatement, TagDbSqlParameters

link_vector = ['https://mysqlclient.readthedocs.io/']
support_matrix = {
    'mysqlclient': {
        '>=3.6': ['2.1.0']
    }
}
note = """"""


def install():
    import wrapt
    import MySQLdb

    _connect = MySQLdb.connect

    def _sw_connect(*args, **kwargs):
        con = _connect(*args, **kwargs)
        con.host = kwargs['host']
        if 'db' in kwargs:
            con.db = kwargs['db']
        else:
            con.db = kwargs['database']
        return ProxyConnection(con)

    class ProxyCursor(wrapt.ObjectProxy):
        def __init__(self, cur):
            wrapt.ObjectProxy.__init__(self, cur)

            self._self_cur = cur

        def __enter__(self):
            return ProxyCursor(wrapt.ObjectProxy.__enter__(self))

        def execute(self, query, args=None):
            peer = f'{self.connection.host}:{self.connection.port}'
            with get_context().new_exit_span(op='Mysql/MysqlClient/execute', peer=peer,
                                             component=Component.MysqlClient) as span:
                span.layer = Layer.Database
                span.tag(TagDbType('mysql'))
                span.tag(TagDbInstance((self.connection.db or '')))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and args:
                    parameter = ','.join([str(arg) for arg in args])
                    max_len = config.sql_parameters_length
                    parameter = f'{parameter[0:max_len]}...' if len(parameter) > max_len else parameter
                    span.tag(TagDbSqlParameters(f'[{parameter}]'))

                return self._self_cur.execute(query, args)

    class ProxyConnection(wrapt.ObjectProxy):
        def __init__(self, conn):
            wrapt.ObjectProxy.__init__(self, conn)

            self._self_conn = conn

        def __enter__(self):
            return ProxyConnection(wrapt.ObjectProxy.__enter__(self))

        def cursor(self, cursorclass=None):
            return ProxyCursor(self._self_conn.cursor(cursorclass))

    MySQLdb.connect = _sw_connect



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpStatusCode

link_vector = ['https://requests.readthedocs.io/en/master/']
support_matrix = {
    'requests': {
        '>=3.6': ['2.26', '2.25']
    }
}
note = """"""


def install():
    from requests import Session

    _request = Session.request

    def _sw_request(this: Session, method, url,
                    params=None, data=None, headers=None, cookies=None, files=None,
                    auth=None, timeout=None, allow_redirects=True, proxies=None,
                    hooks=None, stream=None, verify=None, cert=None, json=None):

        from skywalking.utils.filter import sw_urlparse
        url_param = sw_urlparse(url)

        # ignore trace skywalking self request
        if config.protocol == 'http' and config.collector_address.rstrip('/').endswith(url_param.netloc):
            return _request(this, method, url, params, data, headers, cookies, files, auth, timeout,
                            allow_redirects,
                            proxies,
                            hooks, stream, verify, cert, json)

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else get_context().new_exit_span(op=url_param.path or '/', peer=url_param.netloc,
                                             component=Component.Requests)

        with span:
            carrier = span.inject()
            span.layer = Layer.Http

            if headers is None:
                headers = {}
            for item in carrier:
                headers[item.key] = item.val

            span.tag(TagHttpMethod(method.upper()))
            span.tag(TagHttpURL(url_param.geturl()))

            res = _request(this, method, url, params, data, headers, cookies, files, auth, timeout,
                           allow_redirects,
                           proxies,
                           hooks, stream, verify, cert, json)

            span.tag(TagHttpStatusCode(res.status_code))
            if res.status_code >= 400:
                span.error_occurred = True

            return res

    Session.request = _sw_request



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import time
import traceback
from threading import Thread, Event, current_thread
from typing import Optional

from skywalking import agent
from skywalking import config
from skywalking import profile
from skywalking.loggings import logger
from skywalking.profile.profile_status import ProfileStatusReference, ProfileStatus
from skywalking.profile.profile_task import ProfileTask
from skywalking.profile.snapshot import TracingThreadSnapshot
from skywalking.trace.context import SpanContext
from skywalking.utils.array import AtomicArray
from skywalking.utils.integer import AtomicInteger
from skywalking.utils.time import current_milli_time


class ProfileTaskExecutionContext:
    def __init__(self, task: ProfileTask):
        self.task = task  # type: ProfileTask
        self._current_profiling_cnt = AtomicInteger(var=0)
        self._total_started_profiling_cnt = AtomicInteger(var=0)
        self.profiling_segment_slots = AtomicArray(length=config.profile_max_parallel)
        self._profiling_thread = None  # type: Optional[Thread]
        self._profiling_stop_event = None  # type: Optional[Event]

    def start_profiling(self):
        profile_thread = ProfileThread(self)
        self._profiling_stop_event = Event()

        self._profiling_thread = Thread(target=profile_thread.start, args=[self._profiling_stop_event], daemon=True)
        self._profiling_thread.start()

    def stop_profiling(self):
        if self._profiling_thread is not None and self._profiling_stop_event is not None:
            self._profiling_stop_event.set()

    def attempt_profiling(self, trace_context: SpanContext, segment_id: str, first_span_opname: str) -> \
            ProfileStatusReference:
        """
        check have available slot to profile and add it
        """

        # check has available slot
        using_slot_cnt = self._current_profiling_cnt.get()
        if using_slot_cnt >= config.profile_max_parallel:
            return ProfileStatusReference.create_with_none()

        # check first operation name matches
        if not self.task.first_span_op_name == first_span_opname:
            return ProfileStatusReference.create_with_none()

        # if out limit started profiling count then stop add profiling
        if self._total_started_profiling_cnt.get() > self.task.max_sampling_count:
            return ProfileStatusReference.create_with_none()

        # try to occupy slot
        if not self._current_profiling_cnt.compare_and_set(using_slot_cnt,
                                                           using_slot_cnt + 1):
            return ProfileStatusReference.create_with_none()

        thread_profiler = ThreadProfiler(trace_context=trace_context,
                                         segment_id=segment_id,
                                         profiling_thread=current_thread(),
                                         profile_context=self)

        slot_length = self.profiling_segment_slots.length()
        for idx in range(slot_length):
            # occupy slot success
            if self.profiling_segment_slots.compare_and_set(idx, None, thread_profiler):
                return thread_profiler.profile_status

        return ProfileStatusReference.create_with_none()

    def profiling_recheck(self, trace_context: SpanContext, segment_id: str, first_span_opname: str):
        if trace_context.profile_status.is_being_watched():
            return

        # if first_span_opname was changed by other plugin, there can start profile as well
        trace_context.profile_status.update_status(self.attempt_profiling(trace_context,
                                                                          segment_id,
                                                                          first_span_opname).get())

    def stop_tracing_profile(self, trace_context: SpanContext):
        """
        find tracing context and clear on slot
        """
        for idx, profiler in enumerate(self.profiling_segment_slots):
            if profiler and profiler.matches(trace_context):
                self.profiling_segment_slots.set(idx, None)
                profiler.stop_profiling()
                self._current_profiling_cnt.add_and_get(-1)
                break

    def is_start_profileable(self):
        return self._total_started_profiling_cnt.add_and_get(1) <= self.task.max_sampling_count


class ProfileThread:
    def __init__(self, context: ProfileTaskExecutionContext):
        self._task_execution_context = context
        self._task_execution_service = profile.profile_task_execution_service
        self._stop_event = None  # type: Optional[Event]

    def start(self, stop_event: Event):
        self._stop_event = stop_event

        try:
            self.profiling(self._task_execution_context)
        except Exception as e:
            logger.error('profiling task fail. task_id:[%s] error:[%s]', self._task_execution_context.task.task_id, e)
        finally:
            self._task_execution_service.stop_current_profile_task(self._task_execution_context)

    def profiling(self, context: ProfileTaskExecutionContext):
        max_sleep_period = context.task.thread_dump_period

        while not self._stop_event.is_set():
            current_loop_start_time = current_milli_time()
            profilers = self._task_execution_context.profiling_segment_slots

            for profiler in profilers:  # type: ThreadProfiler
                if profiler is None:
                    continue

                if profiler.profile_status.get() is ProfileStatus.PENDING:
                    profiler.start_profiling_if_need()
                elif profiler.profile_status.get() is ProfileStatus.PROFILING:
                    snapshot = profiler.build_snapshot()
                    if snapshot is not None:
                        agent.add_profiling_snapshot(snapshot)
                    else:
                        # tell execution context current tracing thread dump failed, stop it
                        context.stop_tracing_profile(profiler.trace_context)

            need_sleep = (current_loop_start_time + max_sleep_period) - current_milli_time()
            if not need_sleep > 0:
                need_sleep = max_sleep_period

            # convert to float second
            time.sleep(need_sleep / 1000)


class ThreadProfiler:
    def __init__(self, trace_context: SpanContext, segment_id: str, profiling_thread: Thread,
                 profile_context: ProfileTaskExecutionContext):
        self.trace_context = trace_context
        self._segment_id = segment_id
        self._profiling_thread = profiling_thread
        self._profile_context = profile_context
        self._profile_start_time = -1
        self.profiling_max_time_mills = config.profile_duration * 60 * 1000

        self.dump_sequence = 0

        if trace_context.profile_status is None:
            self.profile_status = ProfileStatusReference.create_with_pending()
        else:
            self.profile_status = trace_context.profile_status  # type: ProfileStatusReference
            self.profile_status.update_status(ProfileStatus.PENDING)

    def start_profiling_if_need(self):
        if current_milli_time() - self.trace_context.create_time > self._profile_context.task.min_duration_threshold:
            self._profile_start_time = current_milli_time()
            self.trace_context.profile_status.update_status(ProfileStatus.PROFILING)

    def stop_profiling(self):
        self.trace_context.profile_status.update_status(ProfileStatus.STOPPED)

    def build_snapshot(self) -> Optional[TracingThreadSnapshot]:
        if not self._profiling_thread.is_alive():
            return None

        current_time = current_milli_time()

        stack_list = []

        # get thread stack of target thread
        stack = sys._current_frames().get(int(self._profiling_thread.ident))
        if not stack:
            return None

        extracted = traceback.extract_stack(stack)
        for idx, item in enumerate(extracted):
            if idx > config.profile_dump_max_stack_depth:
                break

            code_sig = f'{item.filename}.{item.name}: {item.lineno}'
            stack_list.append(code_sig)

        # if is first dump, check is can start profiling
        if self.dump_sequence == 0 and not self._profile_context.is_start_profileable():
            return None

        t = TracingThreadSnapshot(self._profile_context.task.task_id,
                                  self._segment_id,
                                  self.dump_sequence,
                                  current_time,
                                  stack_list)
        self.dump_sequence += 1
        return t

    def matches(self, trace_context: SpanContext) -> bool:
        return self.trace_context == trace_context



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: event/Event.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x65vent/Event.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\x8f\x02\n\x05\x45vent\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12%\n\x06source\x18\x02 \x01(\x0b\x32\x15.skywalking.v3.Source\x12\x0c\n\x04name\x18\x03 \x01(\t\x12!\n\x04type\x18\x04 \x01(\x0e\x32\x13.skywalking.v3.Type\x12\x0f\n\x07message\x18\x05 \x01(\t\x12\x38\n\nparameters\x18\x06 \x03(\x0b\x32$.skywalking.v3.Event.ParametersEntry\x12\x11\n\tstartTime\x18\x07 \x01(\x03\x12\x0f\n\x07\x65ndTime\x18\x08 \x01(\x03\x1a\x31\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"D\n\x06Source\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x17\n\x0fserviceInstance\x18\x02 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x03 \x01(\t*\x1d\n\x04Type\x12\n\n\x06Normal\x10\x00\x12\t\n\x05\x45rror\x10\x01\x32L\n\x0c\x45ventService\x12<\n\x07\x63ollect\x12\x14.skywalking.v3.Event\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x42\x81\x01\n*org.apache.skywalking.apm.network.event.v3P\x01Z1skywalking.apache.org/repo/goapi/collect/event/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')

_TYPE = DESCRIPTOR.enum_types_by_name['Type']
Type = enum_type_wrapper.EnumTypeWrapper(_TYPE)
Normal = 0
Error = 1


_EVENT = DESCRIPTOR.message_types_by_name['Event']
_EVENT_PARAMETERSENTRY = _EVENT.nested_types_by_name['ParametersEntry']
_SOURCE = DESCRIPTOR.message_types_by_name['Source']
Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {

  'ParametersEntry' : _reflection.GeneratedProtocolMessageType('ParametersEntry', (_message.Message,), {
    'DESCRIPTOR' : _EVENT_PARAMETERSENTRY,
    '__module__' : 'event.Event_pb2'
    # @@protoc_insertion_point(class_scope:skywalking.v3.Event.ParametersEntry)
    })
  ,
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'event.Event_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Event)
  })
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.ParametersEntry)

Source = _reflection.GeneratedProtocolMessageType('Source', (_message.Message,), {
  'DESCRIPTOR' : _SOURCE,
  '__module__' : 'event.Event_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.Source)
  })
_sym_db.RegisterMessage(Source)

_EVENTSERVICE = DESCRIPTOR.services_by_name['EventService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n*org.apache.skywalking.apm.network.event.v3P\001Z1skywalking.apache.org/repo/goapi/collect/event/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _EVENT_PARAMETERSENTRY._options = None
  _EVENT_PARAMETERSENTRY._serialized_options = b'8\001'
  _TYPE._serialized_start=401
  _TYPE._serialized_end=430
  _EVENT._serialized_start=58
  _EVENT._serialized_end=329
  _EVENT_PARAMETERSENTRY._serialized_start=280
  _EVENT_PARAMETERSENTRY._serialized_end=329
  _SOURCE._serialized_start=331
  _SOURCE._serialized_end=399
  _EVENTSERVICE._serialized_start=432
  _EVENTSERVICE._serialized_end=508
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List

from skywalking.protocol.profile.Profile_pb2 import ThreadSnapshot, ThreadStack


class TracingThreadSnapshot:

    def __init__(self, task_id: str, trace_segment_id: str, sequence: int, time: int, stack_list: List[str]):
        self.task_id = task_id
        self.trace_segment_id = trace_segment_id
        self.sequence = sequence
        self.time = time
        self.stack_list = stack_list

    def transform(self) -> ThreadSnapshot:
        code_sigs = list(self.stack_list)
        stack = ThreadStack(codeSignatures=code_sigs)

        snapshot = ThreadSnapshot(
            taskId=str(self.task_id),
            traceSegmentId=str(self.trace_segment_id),
            time=int(self.time),
            sequence=int(self.sequence),
            stack=stack
        )

        return snapshot



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import ast
import os

from kafka import KafkaProducer

from skywalking import config
from skywalking.client import ServiceManagementClient, TraceSegmentReportService, LogDataReportService
from skywalking.loggings import logger, logger_debug_enabled
from skywalking.protocol.common.Common_pb2 import KeyStringValuePair
from skywalking.protocol.management.Management_pb2 import InstancePingPkg, InstanceProperties

kafka_configs = {}


def __init_kafka_configs():
    kafka_configs['bootstrap_servers'] = config.kafka_bootstrap_servers.split(',')
    # process all kafka configs in env
    kafka_keys = [key for key in os.environ.keys() if key.startswith('SW_KAFKA_REPORTER_CONFIG_')]
    for kafka_key in kafka_keys:
        key = kafka_key[25:]
        val = os.environ.get(kafka_key)

        if val is not None:
            if val.isnumeric():
                val = int(val)
            elif val in ['True', 'False']:
                val = ast.literal_eval(val)
        else:
            continue

        # check if the key was already set
        if kafka_configs.get(key) is None:
            kafka_configs[key] = val
        else:
            raise KafkaConfigDuplicated(key)


__init_kafka_configs()


class KafkaServiceManagementClient(ServiceManagementClient):
    def __init__(self):
        if logger_debug_enabled:
            logger.debug('kafka reporter configs: %s', kafka_configs)
        self.producer = KafkaProducer(**kafka_configs)
        self.topic_key_register = 'register-'
        self.topic = config.kafka_topic_management

        self.send_instance_props()

    def send_instance_props(self):
        instance = InstanceProperties(
            service=config.service_name,
            serviceInstance=config.service_instance,
            properties=[KeyStringValuePair(key='language', value='Python')],
        )

        key = bytes(self.topic_key_register + instance.serviceInstance, encoding='utf-8')
        value = bytes(instance.SerializeToString())
        self.producer.send(topic=self.topic, key=key, value=value)

    def send_heart_beat(self):
        if logger_debug_enabled:
            logger.debug(
                'service heart beats, [%s], [%s]',
                config.service_name,
                config.service_instance,
            )

        instance_ping_pkg = InstancePingPkg(
            service=config.service_name,
            serviceInstance=config.service_instance,
        )

        key = bytes(instance_ping_pkg.serviceInstance, encoding='utf-8')
        value = bytes(instance_ping_pkg.SerializeToString())
        future = self.producer.send(topic=self.topic, key=key, value=value)
        res = future.get(timeout=10)
        if logger_debug_enabled:
            logger.debug('heartbeat response: %s', res)


class KafkaTraceSegmentReportService(TraceSegmentReportService):
    def __init__(self):
        self.producer = KafkaProducer(**kafka_configs)
        self.topic = config.kafka_topic_segment

    def report(self, generator):
        for segment in generator:
            key = bytes(segment.traceSegmentId, encoding='utf-8')
            value = bytes(segment.SerializeToString())
            self.producer.send(topic=self.topic, key=key, value=value)


class KafkaLogDataReportService(LogDataReportService):
    def __init__(self):
        self.producer = KafkaProducer(**kafka_configs)
        self.topic = config.kafka_topic_log

    def report(self, generator):
        for log_data in generator:
            key = bytes(log_data.traceContext.traceSegmentId, encoding='utf-8')
            value = bytes(log_data.SerializeToString())
            self.producer.send(topic=self.topic, key=key, value=value)


class KafkaConfigDuplicated(Exception):
    def __init__(self, key):
        self.key = key



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: browser/BrowserPerf.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x19\x62rowser/BrowserPerf.proto\x12\rskywalking.v3\x1a\x13\x63ommon/Common.proto\"\xe8\x02\n\x0f\x42rowserPerfData\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x16\n\x0eserviceVersion\x18\x02 \x01(\t\x12\x0c\n\x04time\x18\x03 \x01(\x03\x12\x10\n\x08pagePath\x18\x04 \x01(\t\x12\x14\n\x0credirectTime\x18\x05 \x01(\x05\x12\x0f\n\x07\x64nsTime\x18\x06 \x01(\x05\x12\x10\n\x08ttfbTime\x18\x07 \x01(\x05\x12\x0f\n\x07tcpTime\x18\x08 \x01(\x05\x12\x11\n\ttransTime\x18\t \x01(\x05\x12\x17\n\x0f\x64omAnalysisTime\x18\n \x01(\x05\x12\x0f\n\x07\x66ptTime\x18\x0b \x01(\x05\x12\x14\n\x0c\x64omReadyTime\x18\x0c \x01(\x05\x12\x14\n\x0cloadPageTime\x18\r \x01(\x05\x12\x0f\n\x07resTime\x18\x0e \x01(\x05\x12\x0f\n\x07sslTime\x18\x0f \x01(\x05\x12\x0f\n\x07ttlTime\x18\x10 \x01(\x05\x12\x15\n\rfirstPackTime\x18\x11 \x01(\x05\x12\x0f\n\x07\x66mpTime\x18\x12 \x01(\x05\"\x94\x02\n\x0f\x42rowserErrorLog\x12\x10\n\x08uniqueId\x18\x01 \x01(\t\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x16\n\x0eserviceVersion\x18\x03 \x01(\t\x12\x0c\n\x04time\x18\x04 \x01(\x03\x12\x10\n\x08pagePath\x18\x05 \x01(\t\x12.\n\x08\x63\x61tegory\x18\x06 \x01(\x0e\x32\x1c.skywalking.v3.ErrorCategory\x12\r\n\x05grade\x18\x07 \x01(\t\x12\x0f\n\x07message\x18\x08 \x01(\t\x12\x0c\n\x04line\x18\t \x01(\x05\x12\x0b\n\x03\x63ol\x18\n \x01(\x05\x12\r\n\x05stack\x18\x0b \x01(\t\x12\x10\n\x08\x65rrorUrl\x18\x0c \x01(\t\x12\x1a\n\x12\x66irstReportedError\x18\r \x01(\x08*R\n\rErrorCategory\x12\x08\n\x04\x61jax\x10\x00\x12\x0c\n\x08resource\x10\x01\x12\x07\n\x03vue\x10\x02\x12\x0b\n\x07promise\x10\x03\x12\x06\n\x02js\x10\x04\x12\x0b\n\x07unknown\x10\x05\x32\xb3\x01\n\x12\x42rowserPerfService\x12L\n\x0f\x63ollectPerfData\x12\x1e.skywalking.v3.BrowserPerfData\x1a\x17.skywalking.v3.Commands\"\x00\x12O\n\x10\x63ollectErrorLogs\x12\x1e.skywalking.v3.BrowserErrorLog\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x42\x93\x01\n3org.apache.skywalking.apm.network.language.agent.v3P\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xaa\x02\x1dSkyWalking.NetworkProtocol.V3b\x06proto3')

_ERRORCATEGORY = DESCRIPTOR.enum_types_by_name['ErrorCategory']
ErrorCategory = enum_type_wrapper.EnumTypeWrapper(_ERRORCATEGORY)
ajax = 0
resource = 1
vue = 2
promise = 3
js = 4
unknown = 5


_BROWSERPERFDATA = DESCRIPTOR.message_types_by_name['BrowserPerfData']
_BROWSERERRORLOG = DESCRIPTOR.message_types_by_name['BrowserErrorLog']
BrowserPerfData = _reflection.GeneratedProtocolMessageType('BrowserPerfData', (_message.Message,), {
  'DESCRIPTOR' : _BROWSERPERFDATA,
  '__module__' : 'browser.BrowserPerf_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.BrowserPerfData)
  })
_sym_db.RegisterMessage(BrowserPerfData)

BrowserErrorLog = _reflection.GeneratedProtocolMessageType('BrowserErrorLog', (_message.Message,), {
  'DESCRIPTOR' : _BROWSERERRORLOG,
  '__module__' : 'browser.BrowserPerf_pb2'
  # @@protoc_insertion_point(class_scope:skywalking.v3.BrowserErrorLog)
  })
_sym_db.RegisterMessage(BrowserErrorLog)

_BROWSERPERFSERVICE = DESCRIPTOR.services_by_name['BrowserPerfService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n3org.apache.skywalking.apm.network.language.agent.v3P\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\252\002\035SkyWalking.NetworkProtocol.V3'
  _ERRORCATEGORY._serialized_start=707
  _ERRORCATEGORY._serialized_end=789
  _BROWSERPERFDATA._serialized_start=66
  _BROWSERPERFDATA._serialized_end=426
  _BROWSERERRORLOG._serialized_start=429
  _BROWSERERRORLOG._serialized_end=705
  _BROWSERPERFSERVICE._serialized_start=792
  _BROWSERPERFSERVICE._serialized_end=971
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context, NoopContext
from skywalking.trace.span import NoopSpan
from skywalking.trace.tags import TagHttpMethod, TagHttpURL, TagHttpParams, TagHttpStatusCode, TagHttpStatusMsg

link_vector = ['https://falcon.readthedocs.io/en/stable/']
support_matrix = {
    'hug': {
        '>=3.10': ['2.5', '2.6'],  # api deprecated for 3.10
        '>=3.6': ['2.4.1', '2.5', '2.6'],  # support begins 2.4.1
    }
}
note = """"""


def install():
    from falcon import API, request, RequestOptions

    _original_falcon_api = API.__call__

    def _sw_falcon_api(this: API, env, start_response):
        context = get_context()
        carrier = Carrier()
        req = request.Request(env, RequestOptions())
        headers = req.headers
        method = req.method

        for item in carrier:
            key = item.key.upper()
            if key in headers:
                item.val = headers[key]

        span = NoopSpan(NoopContext()) if config.ignore_http_method_check(method) \
            else context.new_entry_span(op=req.path, carrier=carrier)

        with span:
            span.layer = Layer.Http
            span.component = Component.Falcon
            span.peer = req.remote_addr

            span.tag(TagHttpMethod(method))
            span.tag(TagHttpURL(str(req.url)))

            if req.params:
                span.tag(TagHttpParams(','.join([f'{k}={v}' for k, v in req.params.items()])))

            def _start_response(resp_status, headers):
                try:
                    code, msg = resp_status.split(' ', 1)
                    code = int(code)
                except Exception:
                    code, msg = 500, 'Internal Server Error'

                if code >= 400:
                    span.error_occurred = True

                span.tag(TagHttpStatusCode(code))
                span.tag(TagHttpStatusMsg(msg))

                return start_response(resp_status, headers)

            try:
                return _original_falcon_api(this, env, _start_response)

            except Exception:
                span.raised()

                raise

    API.__call__ = _sw_falcon_api



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" This sub-package is for the convenience of deployment and automation
A CLI for running Python scripts and programs with SkyWalking Python Agent automatically attached.
`loader/sitecustomize.py` is invoked by the Python interpreter at startup.
"""

import logging


def get_cli_logger():
    """ A logger used by sw-python CLI """
    logger = logging.getLogger('skywalking-cli')
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.propagate = False

    return logger


cli_logger = get_cli_logger()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple

import grpc


class _ClientInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
    grpc.StreamUnaryClientInterceptor,
    grpc.StreamStreamClientInterceptor
):

    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(client_call_details, request_iterator, True, False)
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def create(intercept_call):
    return _ClientInterceptor(intercept_call)


ClientCallDetails = namedtuple('ClientCallDetails', ('method', 'timeout', 'metadata', 'credentials'))


def header_adder_interceptor(header, value):
    def intercept_call(client_call_details, request_iterator, request_streaming, response_streaming):
        metadata = list(client_call_details.metadata or ())
        metadata.append((header, value))
        client_call_details = ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata, client_call_details.credentials,
        )
        return client_call_details, request_iterator, None

    return create(intercept_call)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import threading


class AtomicArray:

    def __init__(self, length: int):
        self._length = length
        self._array = [None] * self._length
        self._lock = threading.Lock()

    def __getitem__(self, idx):
        # for iteration
        with self._lock:
            return self._array[idx]

    def length(self) -> int:
        return self._length

    def set(self, idx: int, new_value):
        if idx < 0 or idx >= self.length():
            raise IndexError('atomic array assignment index out of range')

        with self._lock:
            self._array[idx] = new_value

    def get(self, idx: int):
        if idx < 0 or idx >= self.length():
            raise IndexError('atomic array assignment index out of range')

        with self._lock:
            return self._array[idx]

    def compare_and_set(self, idx: int, expect, update) -> bool:
        """
        Atomically sets the value of array to the given updated value if the current value == the expected value
        :return: return True if success, False if the actual value was not equal to the expected value.
        """
        if idx < 0 or idx >= self.length():
            raise IndexError('atomic array assignment index out of range')

        with self._lock:
            if self._array[idx] == expect:
                self._array[idx] = update
                return True

            return False



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
from collections import namedtuple
from enum import Enum
from typing import List


class Component(Enum):
    Unknown = 0
    General = 7000  # built-in modules that may not have a logo to display
    Flask = 7001
    Requests = 7002
    PyMysql = 7003
    Django = 7004
    Tornado = 7005
    Redis = 7
    MongoDB = 9
    PostgreSQL = 22
    KafkaProducer = 40
    KafkaConsumer = 41
    RabbitmqProducer = 52
    RabbitmqConsumer = 53
    Elasticsearch = 47
    Urllib3 = 7006
    Sanic = 7007
    AioHttp = 7008
    Pyramid = 7009
    Psycopg = 7010
    Celery = 7011
    Falcon = 7012
    MysqlClient = 7013
    FastApi = 7014


class Layer(Enum):
    Unknown = 0
    Database = 1
    RPCFramework = 2
    Http = 3
    MQ = 4
    Cache = 5


class Kind(Enum):
    Local = 0
    Entry = 1
    Exit = 2

    @property
    def is_local(self):
        return self == Kind.Local

    @property
    def is_entry(self):
        return self == Kind.Entry

    @property
    def is_exit(self):
        return self == Kind.Exit


LogItem = namedtuple('LogItem', 'key val')


class Log(object):

    def __init__(self, timestamp: time = time.time(), items: List[LogItem] = None): # noqa
        self.timestamp = timestamp
        self.items = items or []



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import traceback
from urllib.parse import urlparse

from skywalking import config


def sw_urlparse(url):
    # Removes basic auth credentials from netloc
    url_param = urlparse(url)
    safe_netloc = url_param.netloc
    try:
        safe_netloc = f"{url_param.hostname}{f':{str(url_param.port)}' if url_param.port else ''}"
    except ValueError:  # illegal url, skip
        pass

    return url_param._replace(netloc=safe_netloc)


def sw_filter(target: str):
    # Remove user:pw from any valid full urls

    return re.sub(r'://(.*?)@', r'://', target)


def sw_traceback():
    stack_trace = traceback.format_exc(limit=config.cause_exception_depth)

    return sw_filter(target=stack_trace)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pathlib

from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / 'README.md').read_text()

setup(
    name='fast-skywalking',
    version='1.6.2',
    description='Python Agent for Fast SkyWalking',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/apache/skywalking-python/',
    author='Apache',
    author_email='dev@skywalking.apache.org',
    license='Apache 2.0',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    install_requires=[
        'grpcio',
        'grpcio-tools',
        'packaging',
        'wrapt',
    ],
    extras_require={
        'test': [
            'testcontainers',
            'pyyaml',
            'pytest',
        ],
        'http': [
            'requests>=2.26.0',
        ],
        'kafka': [
            'kafka',
        ],
    },
    classifiers=[
        'Framework :: Flake8',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: OS Independent',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',

        'Topic :: Software Development',
    ],
    entry_points={
        'console_scripts': [
            'sw-python = skywalking.bootstrap.cli.sw_python:start'
        ]
    },
)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import traceback

from skywalking.log import sw_logging
from skywalking.loggings import logger


def install():
    logger.debug('Installing plugin for logging module')
    # noinspection PyBroadException
    try:
        sw_logging.install()
    except Exception:
        logger.warning('Failed to install sw_logging plugin')
        traceback.print_exc() if logger.isEnabledFor(logging.DEBUG) else None



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component
from skywalking.trace.carrier import Carrier
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagMqBroker, TagMqTopic, TagMqQueue

link_vector = ['https://pika.readthedocs.io']
support_matrix = {
    'pika': {
        '>=3.6': ['1.2']
    }
}
note = """"""


def install():
    from pika.channel import Channel

    _basic_publish = Channel.basic_publish
    __on_deliver = Channel._on_deliver
    Channel.basic_publish = _sw_basic_publish_func(_basic_publish)
    Channel._on_deliver = _sw__on_deliver_func(__on_deliver)


def _sw_basic_publish_func(_basic_publish):
    def _sw_basic_publish(this, exchange,
                          routing_key,
                          body,
                          properties=None,
                          mandatory=False):
        peer = f'{this.connection.params.host}:{this.connection.params.port}'
        context = get_context()
        import pika
        with context.new_exit_span(op=f'RabbitMQ/Topic/{exchange}/Queue/{routing_key}/Producer' or '/',
                                   peer=peer, component=Component.RabbitmqProducer) as span:
            carrier = span.inject()
            span.layer = Layer.MQ
            properties = pika.BasicProperties() if properties is None else properties

            if properties.headers is None:
                properties.headers = {}
            for item in carrier:
                properties.headers[item.key] = item.val

            res = _basic_publish(this, exchange,
                                 routing_key,
                                 body,
                                 properties=properties,
                                 mandatory=mandatory)
            span.tag(TagMqBroker(peer))
            span.tag(TagMqTopic(exchange))
            span.tag(TagMqQueue(routing_key))

            return res

    return _sw_basic_publish


def _sw__on_deliver_func(__on_deliver):
    def _sw__on_deliver(this, method_frame, header_frame, body):
        peer = f'{this.connection.params.host}:{this.connection.params.port}'
        context = get_context()
        exchange = method_frame.method.exchange
        routing_key = method_frame.method.routing_key
        carrier = Carrier()
        for item in carrier:
            try:
                if item.key in header_frame.properties.headers:
                    item.val = header_frame.properties.headers[item.key]
            except TypeError:
                pass

        with context.new_entry_span(op='RabbitMQ/Topic/' + exchange + '/Queue/' + routing_key
                                       + '/Consumer' or '', carrier=carrier) as span:
            span.layer = Layer.MQ
            span.component = Component.RabbitmqConsumer
            __on_deliver(this, method_frame, header_frame, body)
            span.tag(TagMqBroker(peer))
            span.tag(TagMqTopic(exchange))
            span.tag(TagMqQueue(routing_key))

    return _sw__on_deliver



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

""" This module is installed during package setup """
import argparse
import logging

from skywalking.bootstrap import cli_logger
from skywalking.bootstrap.cli import SWRunnerFailure
from skywalking.bootstrap.cli.utility import runner

_options = {
    'run': runner,
}


def start() -> None:
    """ Entry point of CLI """
    parser = argparse.ArgumentParser(description='SkyWalking Python Agent CLI',
                                     epilog='Append your command, with SkyWalking agent attached for you automatically',
                                     allow_abbrev=False)

    parser.add_argument('option', help='CLI options, now only supports `run`, for help please type `sw-python -h` '
                                       'or refer to the CLI documentation',
                        choices=list(_options.keys()))

    # TODO support parsing optional sw_config.yaml
    # parser.add_argument('-config', nargs='?', type=argparse.FileType('r'),
    #                     help='Optionally takes a sw_python.yaml config file')

    # The command arg compress all remaining args into itself
    parser.add_argument('command', help='Your original commands e.g. gunicorn app.wsgi',
                        nargs=argparse.REMAINDER, metavar='command')

    parser.add_argument('-d', '--debug', help='Print CLI debug logs to stdout', action='store_true')

    # To handle cases with flags and positional args in user commands
    args = parser.parse_args()  # type: argparse.Namespace

    cli_logger.setLevel(logging.DEBUG if args.debug else logging.INFO)

    cli_logger.debug(f'Args received {args}')

    if not args.command:
        cli_logger.error('Command is not provided, please type `sw-python -h` for the list of command line arguments')
        return
    try:
        dispatch(args)
    except SWRunnerFailure:
        cli_logger.exception(f"Failed to run the given user application command `{' '.join(args.command)}`, "
                             f'please make sure given command is valid.')
        return


def dispatch(args: argparse.Namespace) -> None:
    """ Dispatches parsed args to a worker """
    cli_option, actual_command = args.option, args.command

    cli_logger.debug(f"SkyWalking Python agent with CLI option '{cli_option}' and command {actual_command}")

    # Dispatch actual user application command to runner
    _options[cli_option].execute(actual_command)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: browser/BrowserPerfCompat.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..common import Common_pb2 as common_dot_Common__pb2
from ..browser import BrowserPerf_pb2 as browser_dot_BrowserPerf__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1f\x62rowser/BrowserPerfCompat.proto\x1a\x13\x63ommon/Common.proto\x1a\x19\x62rowser/BrowserPerf.proto2\xb3\x01\n\x12\x42rowserPerfService\x12L\n\x0f\x63ollectPerfData\x12\x1e.skywalking.v3.BrowserPerfData\x1a\x17.skywalking.v3.Commands\"\x00\x12O\n\x10\x63ollectErrorLogs\x12\x1e.skywalking.v3.BrowserErrorLog\x1a\x17.skywalking.v3.Commands\"\x00(\x01\x42\xa4\x01\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\x01Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\xb8\x01\x01\xaa\x02$SkyWalking.NetworkProtocol.V3.Compatb\x06proto3')



_BROWSERPERFSERVICE = DESCRIPTOR.services_by_name['BrowserPerfService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n:org.apache.skywalking.apm.network.language.agent.v3.compatP\001Z:skywalking.apache.org/repo/goapi/collect/language/agent/v3\270\001\001\252\002$SkyWalking.NetworkProtocol.V3.Compat'
  _BROWSERPERFSERVICE._serialized_start=84
  _BROWSERPERFSERVICE._serialized_end=263
# @@protoc_insertion_point(module_scope)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from skywalking import Layer, Component, config
from skywalking.trace.context import get_context
from skywalking.trace.tags import TagDbType, TagDbInstance, TagDbStatement, TagDbSqlParameters

link_vector = ['https://www.psycopg.org/']
support_matrix = {
    'psycopg2-binary': {
        '>=3.10': [],
        '>=3.6': ['2.9']  # transition to psycopg(3), not working for python 3.10
    }
}
note = """"""


def install():
    import wrapt  # psycopg2 is read-only C extension objects so they need to be proxied
    import psycopg2

    class ProxyCursor(wrapt.ObjectProxy):
        def __init__(self, cur):
            wrapt.ObjectProxy.__init__(self, cur)

            self._self_cur = cur

        def __enter__(self):
            return ProxyCursor(wrapt.ObjectProxy.__enter__(self))

        def execute(self, query, vars=None):
            dsn = self.connection.get_dsn_parameters()
            peer = f"{dsn['host']}:{dsn['port']}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/execute', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length and vars is not None:
                    text = ','.join(str(v) for v in vars)

                    if len(text) > config.sql_parameters_length:
                        text = f'{text[:config.sql_parameters_length]}...'

                    span.tag(TagDbSqlParameters(f'[{text}]'))

                return self._self_cur.execute(query, vars)

        def executemany(self, query, vars_list):
            dsn = self.connection.get_dsn_parameters()
            peer = f"{dsn['host']}:{dsn['port']}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/executemany', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(query))

                if config.sql_parameters_length:
                    max_len = config.sql_parameters_length
                    total_len = 0
                    text_list = []

                    for vars in vars_list:
                        text = f"[{','.join(str(v) for v in vars)}]"
                        total_len += len(text)

                        if total_len > max_len:
                            text_list.append(f'{text[:max_len - total_len]}...')

                            break

                        text_list.append(text)

                    span.tag(TagDbSqlParameters(f"[{','.join(text_list)}]"))

                return self._self_cur.executemany(query, vars_list)

        def callproc(self, procname, parameters=None):
            dsn = self.connection.get_dsn_parameters()
            peer = f"{dsn['host']}:{dsn['port']}"

            with get_context().new_exit_span(op='PostgreSLQ/Psycopg/callproc', peer=peer,
                                             component=Component.Psycopg) as span:
                span.layer = Layer.Database
                args = f"({'' if not parameters else ','.join(parameters)})"

                span.tag(TagDbType('PostgreSQL'))
                span.tag(TagDbInstance(dsn['dbname']))
                span.tag(TagDbStatement(procname + args))

                return self._self_cur.callproc(procname, parameters)

    class ProxyConnection(wrapt.ObjectProxy):
        def __init__(self, conn):
            wrapt.ObjectProxy.__init__(self, conn)

            self._self_conn = conn

        def __enter__(self):
            return ProxyConnection(wrapt.ObjectProxy.__enter__(self))

        def cursor(self, *args, **kwargs):
            return ProxyCursor(self._self_conn.cursor(*args, **kwargs))

    def connect(*args, **kwargs):
        return ProxyConnection(_connect(*args, **kwargs))

    _connect = psycopg2.connect
    psycopg2.connect = connect
